function order_pairs(tab)
	local sorted = {}
	for key in pairs(tab) do
		table.insert(sorted,key)
	end
	table.sort(sorted)
	local i=0
	return function()
		i = i + 1
		if i > #sorted then
			return nil,nil
		else
			local key=sorted[i]
			return key,tab[key]
		end
	end
end

--行頭のTRコマンドで指定されたトラックの配列を返す
function get_line_tracks(line_text)
	local tracks = {}
	local find_st,find_end,num_st = string.find(line_text, "TR([%d,]+)")
	--TR??が無い場合は終了
	if (not find_st) then
		--print('TRコマンドが見つかりません。')
		--print(line_text)
		return nil
	end
	
	--コンマ区切りの数字列を抽出
	for tr in string.gmatch(num_st,"%d+") do
		table.insert(tracks, tonumber(tr))
	end
	return tracks
end

--ファイル先頭からend_line行まで#timebaseコマンドを探して返す
function find_timebase_set(end_line)
	local timebase = 24		--見つからなければデフォルト値
	for now_line = 0,end_line do
		local line = GetParagraphText(-1,now_line)
		line = string.lower(line)
		local cap = string.match(line,"#timebase%s+(%d+)")
		if (cap ~= nil) then
			timebase = tonumber(cap)
		end
	end
	return timebase
end

--コメントを削除
function erase_comment(string)
	local out_str
	out_str = string.gsub(string,"/%*.-%*/", "")
	out_str = string.gsub(out_str,"//.-\r", "\r")
	out_str = string.gsub(out_str,"//.-$", "")
	out_str = string.gsub(out_str,";.-\r", "\r")
	out_str = string.gsub(out_str,";.-$", "")
	return out_str
end

--コメントを削除（ただし、;@はコメントとみなさない
function erase_comment_except_hacks(string)
	local out_str
	out_str = string.gsub(string,"/%*.-%*/", "")
	out_str = string.gsub(out_str,"//.-\r", "\r")
	out_str = string.gsub(out_str,"//.-$", "")
	out_str = string.gsub(out_str,";[^@].-\r", "\r")
	out_str = string.gsub(out_str,";[^@].-$", "")
	return out_str
end

--;@を削除
function erase_hacks(string)
	local out_str
	out_str = string.gsub(string,";@.-\r", "\r")
	out_str = string.gsub(out_str,";@.-$", "")
	return out_str
end

--num番のサブルーチンを探し、見つかればその内容を返す
function find_sub(num)
	local end_line = GetParagraphCount(-1) - 1
	local last_line = GetParagraphRange(-1,end_line)
	local whole_text = GetText(-1,0,last_line.e)
	local ptn = string.format("[sS][uU][bB]%%(%%s*%d%%s*%%)%%s*{(.-)}",num)
	local sub_cont = string.match(whole_text,ptn)
	if (not sub_cont) then
		return ""
	end
	return sub_cont
end

--$nameのマクロ定義を探し、見つかればその内容を返す
function find_macro(name)
	local end_line = GetParagraphCount(-1) - 1
	local last_line = GetParagraphRange(-1,end_line)
	local whole_text = GetText(-1,0,last_line.e)
	local ptn = string.format("$%s%%s*{(.-)}",name)
	local sub_cont = string.match(whole_text,ptn)
	if (not sub_cont) then
		return ""
	end
	return sub_cont
end

--line行からファイル先頭に向かってBGM定義を探し、最初に見つかったBGMブロックの内容を返す
function find_bgm_block(line)
	local bgm_block = {}
	--lineがBGMブロックの中にあれば、その開始位置と終了位置を求める
	local now_line = line
	local bgm_line = -1
	local block_find_st
	local line_text
	local find_st,find_end
	--BGM定義の行を探す
	while (bgm_line == -1) and (now_line >= 0) do
		line_text = GetParagraphText(-1,now_line)
		line_text = string.lower(line_text)
		find_st,find_end = string.find(line_text, "bgm%(%d*%)")
		if find_st ~= nil then
			bgm_line = now_line
			block_find_st = find_end+1
		end
		now_line = now_line - 1
	end
	--先頭まで探して見つからなければ終了
	if bgm_line == -1 then
		print('BGMブロックの始点が見つかりません.')
		return nil
	end
	--ブロック開始位置を探す
	now_line = bgm_line
	bgm_block.s = -1
	bgm_block.sl = -1
	while (now_line <= line) and (bgm_block.s == -1) do
		line_text = GetParagraphText(-1,now_line)
		line_text = string.gsub(line_text,"K{[a-zA-Z%-]+}","") --Kコマンドを無視
		find_st,find_end = string.find(line_text, "{", block_find_st)
		if find_st ~= nil then
			bgm_block.sl = now_line
			bgm_block.si = find_end + 1
			bgm_block.s = GetParagraphStartTextIndex(-1,now_line) + bgm_block.si - 1
		end
		now_line = now_line + 1
		block_find_st = 0
	end
	--カーソル位置まで探して見つからなければ終了
	if bgm_block.s == -1 then
		print('BGMブロックの始点が見つかりません.')
		return nil
	end
	--ブロック終了位置を探す
	block_find_st = find_end+1
	now_line = now_line - 1
	local paragraph_count = GetParagraphCount(-1)
	bgm_block.e = -1
	while (now_line <= paragraph_count) and (bgm_block.e == -1) do
		line_text = GetParagraphText(-1,now_line)
		line_text = string.gsub(line_text,"K{[a-zA-Z%-]+}","") --Kコマンドを除去
		find_st,find_end = string.find(line_text, "}", block_find_st)
		if find_st ~= nil then
			bgm_block.el = now_line
			bgm_block.ei = find_st - 1
			bgm_block.e = GetParagraphStartTextIndex(-1,now_line) + bgm_block.ei
		end
		now_line = now_line + 1
		block_find_st = 0
	end
	--終端まで探して見つからなければ終了
	if bgm_block.e == -1 then
		print('BGMブロックの終端が見つかりません.')
		return nil
	end
	
	return bgm_block
end

--テーブルtabl内に要素valの出現回数を求める
function find_in_table(tabl,val)
	local found = 0
	for ind=1,#tabl do
		if (tabl[ind] == val) then
			found = found + 1
		end
	end
	return found
end

--文字列strで表されたレングスの合計値を返す
function calc_len_total(str)
	--%,数値,+,-で構成された長さ情報を合算する
	--print(str)
	local tick = 0
	for val,dot in string.gmatch(str,"([%+%-]?%%?%d+)(%.*)") do
		--print(dot)
		--print(val)
		val = string.gsub(val,"^%+","") --'+'を除去
		if (string.sub(val,1,2) == '-%') then
			tick = tick - tonumber(string.sub(val,3)) * (2-2^(-#dot))
		elseif (string.sub(val,1,1) == '%') then
			tick = tick + tonumber(string.sub(val,2)) * (2-2^(-#dot))
		else
			tick = tick + ((timebase * 4) / tonumber(val)) * (2-2^(-#dot))
		end
	end
	return tick
end

--100*分母+分子で表された数値データを分子、分母の数値に分ける
function decode_time_representation(time)
	local num = math.floor(time / 100)
	local denom = time - num * 100
	return num,denom
end

--分子、分母を100*分母+分子の表現に変換する
function encode_time_representation(num,denom)
	return (num * 100 + denom)
end

--拍子変更位置テーブル(key=変更tick,value=拍子)を元に、
--実際に拍子が変更される小節とTick値を求める
--※拍子変更が、小節の区切り以外の位置にあった場合、次の小節で拍子が変更される
function time_change_tick2workbar(time_change_tick,offset)
	--何も無ければ1.1.0に4/4があることにする
	local beat_num = 4
	local beat_denom = 4
	local bar_tick = (timebase * beat_num * 4) / beat_denom
	
	--先頭からoffset[Tick]の位置を1.1.0とする
	local bar_change_work_tick = offset
	local time_change_work_bar = 1
	
	local bar_change_bar = {}
	local time_change_bar = {}
	for tick,time in order_pairs(time_change_tick) do
		--print(tick)
		--print(time)
		local	next_bar_change_tick = tick
		local	bar_length = math.ceil((next_bar_change_tick - bar_change_work_tick) / bar_tick)
		--print(bar_tick)
		--print(next_bar_change_tick)
		--print(bar_change_work_tick)
		--print(bar_length)
		bar_change_work_tick = bar_change_work_tick + bar_length * bar_tick
		--print(bar_change_work_tick)
		time_change_work_bar = time_change_work_bar + bar_length
		--print(time_change_work_bar)
		table.insert(bar_change_bar,time_change_work_bar,bar_change_work_tick)
		table.insert(time_change_bar,time_change_work_bar,time)
		
		beat_num,beat_denom = decode_time_representation(time)
		bar_tick = (timebase * beat_num * 4) / beat_denom
	end
	return time_change_bar,bar_change_bar
end

--in_bar小節、in_bar拍、in_clkティックの情報を、先頭からのTick数に変換する
--time_changeテーブル(key=変更tick,value=拍子)の拍子変更情報を使用する
--先頭から、offsetティックの位置を 1.1.0として計算する
function bbc2tick(in_bar,in_beat,in_clk,time_change,offset)
	local bar_time,bar_tick = time_change_tick2workbar(time_change,offset)
	local beat_num = 4
	local beat_denom = 4
	local ctime_begin_tick = offset
	local ctime_change_begin_bar = 1
	for bar,tick in order_pairs(bar_tick) do
		if (bar >= in_bar) then
			break
		end
		beat_num,beat_denom = decode_time_representation(bar_time[bar])
		ctime_begin_tick = tick
		ctime_change_begin_bar = bar
	end
	local beat_clk = (timebase * 4) / beat_denom
	local bar_clk = beat_clk * beat_num
	return (ctime_begin_tick + (in_bar - ctime_change_begin_bar) * bar_clk + (in_beat - 1) * beat_clk + in_clk)
end

--先頭からin_tickの位置を、（小節、拍、ティック)に変換する
--time_changeテーブル(key=変更tick,value=拍子)の拍子変更情報を使用する
--先頭から、offsetティックの位置を 1.1.0として計算する
function tick2bbc(in_tick,time_change,offset)
	local bar_time,bar_tick = time_change_tick2workbar(time_change,offset)
	local beat_num = 4
	local beat_denom = 4
	local ctime_begin_tick = offset
	local ctime_change_begin_bar = 1
	for bar,tick in order_pairs(bar_tick) do
		--print(bar)
		--print(tick)
		if (tick >= in_tick) then
			break
		end
		beat_num,beat_denom = decode_time_representation(bar_time[bar])
		ctime_begin_tick = tick
		ctime_change_begin_bar = bar
	end
	--print(ctime_begin_tick)
	local beat_clk = (timebase * 4) / beat_denom
	local bar_clk = beat_clk * beat_num
	local out_bar = ctime_change_begin_bar + math.floor((in_tick - ctime_begin_tick) / bar_clk)
	local out_beat = math.floor((in_tick - ctime_begin_tick) / beat_clk) % beat_num + 1
	local out_clk = (in_tick - ctime_begin_tick) % beat_clk
	return out_bar,out_beat,out_clk
end

--マクロ名vのマクロを探して内容を返す
function extract_macro(v)
	--print(v)
	if (not macro_table[v]) then
		macro_table[v] = find_macro(v)
	end
	return macro_table[v]
end

--番号vのサブルーチンを探して内容を返す
function extract_subrotine(v)
	--print(v)
	if (not subroutine_table[v]) then
		subroutine_table[v] = find_sub(v)
	end
	return subroutine_table[v]
end
		
--リピート(A)パターン置換用
function extract_repeat_A(rep1,cnt_a,cnt_b,rep2)
	if (#rep1 > 0) and (#rep2 > 0) then
		 print("リピート回数が両方に記述されています.")
		 return
	end
	if (#rep1 == 0) and (#rep2 == 0) then
		 print("リピート回数が記述されていません.")
		 return
	end
	local repnum
	if ( #rep1 > 0 ) then
		repnum = tonumber(rep1)
	end
	if ( #rep2 > 0 ) then
		repnum = tonumber(rep2)
	end
	local conv_out = string.rep(cnt_a..cnt_b,repnum-1)
	return conv_out..cnt_a
end

--リピート(B)パターン置換用
function extract_repeat_B(cnt_a,cnt_b)
	local conv_out
	if #cnt_a == 0 then
		conv_out = "|:"..cnt_b..cnt_b
	else
		conv_out = "|:"..cnt_a..cnt_b..cnt_a
	end
	founds = true
	return conv_out
end

--関数ここまで--


--カーソル位置を取得
selected_pos=GetSelectedRange(-1)

--選択状態がある場合は、末尾位置をカーソル位置とする
if selected_pos.s ~= selected_pos.e then
	selected_pos.s = selected_pos.e
	SetSelectedRange(-1,selected_pos.s,selected_pos.e)
end

--直近のBGMブロックを探す

--カーソルがあるトラックをカウントする
curs_line = GetParagraphIndexFromTextIndex(-1,selected_pos.s)
target_tracks = get_line_tracks(GetParagraphText(-1,curs_line))
if (not target_tracks) then
	print('TRコマンドのある行で実行して下さい.')
	return
end

--timebase設定を取得
timebase = find_timebase_set(curs_line)
defalut_len = timebase
--print(timebase)

macro_table = {}
subroutine_table = {}
total_tick_table = {}

begin_tick = {}
time_change_table = {}

for i=1,#target_tracks do
	--begin_tick[target_tracks[i]] = 0
	time_change_table[target_tracks[i]] = {}

	--直近のBGMブロックを探す
	local block = find_bgm_block(curs_line)
	--print(block.sl)
	--print(block.el)
	if (block.s ~= nil) and (block.e >= selected_pos.s) then
		local req_text = GetText(-1,block.s,selected_pos.s)
		--コメントを取り除く
		req_text = erase_comment_except_hacks(req_text)
		--print(req_text)
		--1行ずつ取り出す
		local	req_line = {}
		local	start_pos = ""
		local	change_time = ""
		for line in string.gmatch(req_text, "[^\r]+") do
			--print(line)
			--基準位置指定があれば、次のTRコマンドの行を0tick目にする（各トラック最初の指定のみ有効）
			local bar,beat,clk = string.match(line,";@(%d+)%.(%d+)%.(%d+)")
			--拍子指定があれば、次のTRコマンドの行にコマンを埋め込む
			local numerator,denominator = string.match(line,";@(%d+)/(%d+)")
			line = erase_hacks(line)
			--print(line)
			
			--調べたいトラックの行のみを取り出す
			local line_tracks = get_line_tracks(line)
			local found = 0
			if line_tracks ~= nil then
				found = find_in_table(line_tracks,target_tracks[i])
				--開始位置、拍子指定コマンドを埋め込む
				if (found > 0) then
					if (#start_pos > 0) then
						table.insert( req_line,"TR"..table.concat(line_tracks,",")..start_pos )
						start_pos = ""
					end
					if (#change_time > 0) then
						table.insert( req_line,"TR"..table.concat(line_tracks,",")..change_time )
						change_time = ""
					end
				else
					start_pos = ""
					change_time = ""
				end
			end
			if found > 0 then
				table.insert(req_line,line)
				--print(line)
			end
			
			--拍子変更、拍指定コマンドを生成
			if (bar ~= nil) then
				start_pos = string.format(" l=%s+%s+%s ",bar,beat,clk)
				--print(start_pos)
			end
			if (numerator ~= nil) then
				change_time = string.format(" l#%s+%s ",numerator,denominator)
				--print(change_time)
			end
		end
		--行を結合する
		req_text = table.concat(req_line)
		--print(req_text)
		
		--マクロの展開
		req_text = string.gsub(req_text,"$([^%c%s]+)",extract_macro)
		--print(req_text)
		
		--リピート(A)の展開
		req_text = string.gsub(req_text,"%[(%d*)([^:%]]*):?([^:%]]*)%](%d*)",extract_repeat_A)
		--print(req_text)
		
		--リピート(B)の展開
		local founds=false
		repeat
			founds=false
			req_text = string.gsub(req_text,"%|:([^%\\:]-)%\\?([^%\\:]-):%|",extract_repeat_B)
			--print(req_text)
		until founds == false
		req_text = string.gsub(req_text,"%|:","") --残った記号を除去
		--print(req_text)
		
		--サブルーチンの展開
		req_text = string.gsub(req_text,"S(%d+)",extract_subrotine)
		--print(req_text)
		
		--マクロ、サブルーチン内の不要部分を削除
		req_text = erase_comment(req_text)
		req_text = string.gsub(req_text,"Scale%(.-%)","")
		req_text = string.gsub(req_text,"K{.-}","")
		
		--print(req_text)
		
		--長さのあるコマンドの抽出
		local abs_tick = 0
		interp_note = function(note_sig,note_time)
			--print(note_sig..";"..note_time)
			if ( note_sig == "l=" ) then
				--絶対位置指定コマンド
				--print(note_sig..";"..note_time)
				local	bar,beat,clk = string.match(note_time,"(%d+)%+(%d+)%+(%d+)")
				if ( not bar ) then
					print("l=は使用出来ません.")
				else
					--トラック中最初の指定だけが有効
					if ( begin_tick[target_tracks[i]] == nil ) then
						bar = tonumber(bar)
						beat = tonumber(beat)
						clk = tonumber(clk)
						local	written_tick = bbc2tick(bar,beat,clk,time_change_table[target_tracks[i]],0)
						--現在位置と指定位置との差を開始tickにする
						--print(written_tick)
						--print(abs_tick)
						begin_tick[target_tracks[i]] = abs_tick - written_tick
					end
				end
			elseif (note_sig == "l#" ) then
				--拍子変更コマンド
				--print(note_sig..";"..note_time)
				local numerator,denominator = string.match(note_time,"(%d+)%+(%d+)")
				if ( not numerator ) then
					print("l#は使用出来ません.")
				else
					table.insert(time_change_table[target_tracks[i]],abs_tick,encode_time_representation(numerator,denominator))
				end
			else
				local cmd = string.sub(note_sig,1,1)
				if (cmd ~= "l") then
					--音長が省略されたノート、休符、タイの直後に、デフォルト音長を付加
					local	initial_char = string.match(note_time,"^[%d%%]")
					if (initial_char == nil) then
						--print(initial_char)
						note_time = "%" .. tostring(defalut_len) .. note_time
						--print(note_time)
					end
				end
				--print(note_sig..";"..note_time)
				local note_tick_total = calc_len_total(note_time)
				if ( note_tick_total <= 0 ) or ((math.ceil(note_tick_total) - note_tick_total) > 0.01) then
					print( "\"" .. note_sig .. note_time .. "\": 不正な音長です.")
				end
				
				if cmd == "l" then
					--lコマンドの場合は加算しない
					--print(note_tick_total)
					--print("l")
					if ( note_sig == "l" ) then
						defalut_len = note_tick_total
					end
				else
					--print(note_tick_total)
					--print(";")
					abs_tick = abs_tick + note_tick_total
				end
			end
		end
		req_text = string.gsub(req_text,"([a-grl^][%+#%-=]?)([%d%%%+%-%.]*)",interp_note)
		
		total_tick_table[ target_tracks[i] ] = abs_tick
	end
end

--[[
print(begin_tick[4])
bar_times,bar_ticks = time_change_tick2workbar(time_change_table[4],begin_tick[4])
for tick,time in order_pairs(bar_ticks) do
	print(tick)
	print(time)
end
--]]

out_text = {}
for tr,tick in pairs(total_tick_table) do
	--位置指定されていなければ、先頭からのTick数を出力
	if ( not begin_tick[tr] ) then
		table.insert(out_text,string.format("TR%d:%d",tr,tick))
	else
		local	bar,beat,clk = tick2bbc(tick,time_change_table[tr],begin_tick[tr])
		table.insert(out_text,string.format("TR%d:%d.%d.%d",tr,bar,beat,clk))
	end
end
if #out_text > 0 then
	table.insert(out_text,1,"/*")
	table.insert(out_text,"*/")
	InputText(-1,table.concat(out_text," ") )
end
